/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Thumbs Down.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


const THUMBSDOWN_TOPIC_DB_CHANGED = "thumbsdown-db-changed";
const THUMBSDOWN_CSS_SELECTED = "thumbsdown_selected";


var gThumbsDown =
{
    onLoad: function(e)
    {
        // util.js
        this.util = {};
        var loader = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
        loader.loadSubScript("chrome://thumbsdown/content/util.js", this.util);
        
        // localization
        this.util.stringBundle = this.util.$("thumbsdown_stringBundle");
        
        // database connection
        this.sqlExec = new thumbsdown_SqlExecutor(THUMBSDOWN_SQLITE_FILENAME, THUMBSDOWN_SQL_EXECUTOR_CONFIG, THUMBSDOWN_SQL_EXECUTOR_INIT_CONFIG);
        
        // init shortcut keys
        this.util.initKey("thumbsdown_key_toggleStatus",         "extensions.thumbsdown.keys.toggleStatus");
        this.util.initKey("thumbsdown_key_activateAndSelectAll", "extensions.thumbsdown.keys.activateAndSelectAll");
        this.util.initKey("thumbsdown_key_download",             "extensions.thumbsdown.keys.download");
        
        // stylesheet
        var sss = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
        var ios = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        var uri = ios.newURI("chrome://thumbsdown/content/thumbsdown.css", null, null);
        if (! sss.sheetRegistered(uri, sss.USER_SHEET))
            sss.loadAndRegisterSheet(uri, sss.USER_SHEET);
        
        // script change (database change) observer
        var obsSvc = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        obsSvc.addObserver(this, THUMBSDOWN_TOPIC_DB_CHANGED, false);
        
        // preference observers
        var prefSvc = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService);
        this.prefs = prefSvc.getBranch("extensions.thumbsdown.");
        this.prefs.QueryInterface(Ci.nsIPrefBranch2);
        this.prefs.addObserver("showInStatusBar", this, false);
        
        // listeners
        gBrowser.tabContainer.addEventListener("TabSelect", function(e) { gThumbsDown.onTabSelect(e) }, false);
        gBrowser.tabContainer.addEventListener("TabClose", function(e) { gThumbsDown.onTabClose(e) }, false);
        gBrowser.addEventListener("DOMContentLoaded", function(e) { gThumbsDown.onPageLoad(e); }, false);
        gBrowser.addEventListener("click", function(e) { gThumbsDown.onContentClick(e); }, true);
        
        // popup menu listeners
        this.util.$("menu_ToolsPopup").addEventListener("popupshowing", function(e) { gThumbsDown.onToolsMenuShowing(e); }, false);
        this.util.$("contentAreaContextMenu").addEventListener("popupshowing", function(e) { gThumbsDown.onContextMenuShowing(e); }, false);
        
        // init ui
        var ctx = new ThumbsDownContext();
        this._updateUi(ctx, true);
    },
    
    onUnload: function(e)
    {
        // preference observers
        this.prefs.removeObserver("showInStatusBar", this);
        
        // observers
        var obsSvc = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        obsSvc.removeObserver(this, THUMBSDOWN_TOPIC_DB_CHANGED);
    },
    
    observe: function(subject, topic, data)
    {
        switch (topic)
        {
            case THUMBSDOWN_TOPIC_DB_CHANGED:
                this.scripts = null;
                break;
            case "nsPref:changed":
                this.observePreference(subject, topic, data)
                break;
            default:
                throw new Error("unknown topic: " + topic);
        }
    },
    
    observePreference: function(subject, topic, data)
    {
        switch (data)
        {
            case "showInStatusBar":
                var ctx = this._getContext(null, true, false);
                this._updateUi(ctx, true);
                break;
            default:
                throw new Error("unknown preference: " + data);
        }
    },
    
    onToolsMenuShowing: function(e)
    {
        var show = nsPreferences.getBoolPref("extensions.thumbsdown.showInToolsMenu");
        this.util.$("thumbsdown_toolsMenu_openScriptRunner").hidden = (! show);
    },
    
    onContextMenuShowing: function(e)
    {
        var show = nsPreferences.getBoolPref("extensions.thumbsdown.showInContextMenu");
        this.util.$("thumbsdown_contextMenu_menu").hidden = (! show);
    },
    
    onTabSelect: function(e)
    {
        var ctx = this._getContext(null, true, false);
        this._updateUi(ctx);
    },
    
    onTabClose: function(e)
    {
        var browser = e.target.linkedBrowser;
        var ctx = this._getContext(browser, false, false);
        if (! ctx)
            return;
        ctx.isActive = false;
    },
    
    onPageLoad: function(e)
    {
        var doc = e.originalTarget;
        if (! (doc instanceof HTMLDocument))
            return;
        if (doc.defaultView.frameElement)
            return;
        var browser = this._getBrowserForDocument(doc);
        var ctx = this._getContext(browser, false, false);
        if (! ctx)
            return;
        ctx.isActive = false;
        if (browser != gBrowser.selectedBrowser)
            return;
        this._updateUi(ctx);
    },
    
    onContentClick: function(e)
    {
        if (! (e.button == 0))
            return;
        var elem = e.originalTarget;
        if (! ((elem instanceof HTMLImageElement) &&  (elem.parentNode instanceof HTMLAnchorElement)))
            return;
        var ctx = this._getContext(null, false, false);
        if (! (ctx && ctx.isActive))
            return;
        var changed = ctx.toggleThumbnail(elem);
        e.stopPropagation();
        e.preventDefault();
        if (! changed)
            return;
        this._updateUi(ctx);
    },
    
    onStatusBarClick: function(e)
    {
        // single click
        if      ((e.button == 0) && (e.detail == 1))
            this.toggleStatus();
        // double click
        else if ((e.button == 0) && (e.detail == 2))
        {
            this.activate();
            this.selectAll();
        }
        // single middle click
        else if ((e.button == 1) && (e.detail == 1))
            this.download();
    },
    
    toggleStatus: function()
    {
        var ctx = this._getContext(null, false, false);
        if (ctx && ctx.isActive)
            this.deactivate();
        else
            this.activate();
    },
    
    activate: function()
    {
        var browser = gBrowser.selectedBrowser;
        var doc = browser.contentDocument;
        // TODO: better valid doc check (image is HTMLDocument)
        if (! (doc instanceof HTMLDocument))
            return;
        var ctx = this._getContext(null, true, true);
        if (ctx.isActive)
            return;
        if (! this._checkSettings())
            return;
        ctx.isActive = true;
        this._updateUi(ctx);
    },
    
    deactivate: function()
    {
        var ctx = this._getContext(null, false, false);
        if (! ctx)
            return;
        ctx.isActive = false;
        this._updateUi(ctx);
    },
    
    selectAll: function()
    {
        var ctx = this._getContext(null, false, false);
        if (! (ctx && ctx.isActive))
            return;
        var images = gBrowser.selectedBrowser.contentDocument.images;
        var changed = ctx.selectThumbnails(images);
        if (changed)
            this._updateUi(ctx);
    },
    
    download: function()
    {
        var ctx = this._getContext(null, false, false);
        if (! (ctx && ctx.isActive))
            return;
        var browser = gBrowser.selectedBrowser;
        var docUrl = browser.contentDocument.location.href;
        var urls = [];
        var scriptArgs = [];
        for each (image in ctx.thumbnails)
        {
            var script = image.thumbsdown;
            switch (script.type)
            {
                // direct
                case 0:
                    var url = {};
                    url.url = image.parentNode.href;
                    url.referer = docUrl;
                    urls.push(url);
                    break;
                // regexp substitution
                case 1:
                    var url = {};
                    var sourceUrl = (script.isSourceThumbnail ? image.src : image.parentNode.href);
                    url.url = sourceUrl.replace(script.pattern, script.template);
                    url.referer = image.parentNode.href;
                    urls.push(url);
                    break;
                // script
                case 2:
                    var args = {};
                    args.name = script.name;
                    args.code = script.script;
                    args.documentUrl = docUrl;
                    args.thumbnailUrl = image.src;
                    args.linkUrl = image.parentNode.href;
                    scriptArgs.push(args);
                    break;
                default:
                    throw new Error("invalid type: " + entry.type);
            }
        }
        var dir = nsPreferences.copyUnicharPref("extensions.thumbsdown.downloadDir");
        if (urls.length > 0)
            this._downloadUrls(urls, dir);
        if (scriptArgs.length > 0)
            this._runScripts(scriptArgs, dir);
        ctx.isActive = false;
        this._updateUi(ctx);
    },
    
    _downloadUrls: function(urls, dir)
    {
        try
        {
            var dmSvc = Cc["@thumbsdown.mozdev.org/download-manager-service;1"].getService(Ci.nsISupports).wrappedJSObject;
            dmSvc.download(urls, dir);
        }
        catch (e)
        {
            var promptSvc = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
            var title = this.util._("error.title");
            var msg = this.util._("error.download", [e.message]);
            promptSvc.alert(window, title, msg);
        }
    },
    
    _runScripts: function(scriptArgs, dir)
    {
        var scriptRunner = Cc["@thumbsdown.mozdev.org/script-runner;1"].getService(Ci.nsISupports).wrappedJSObject;
        var callback = this._getScriptCallback(dir);
        scriptArgs.forEach(function(args) { args.callback = callback });
        scriptRunner.addScripts(scriptArgs);
        
        if (nsPreferences.getBoolPref("extensions.thumbsdown.scriptrunner.open"))
        {
            var focus = nsPreferences.getBoolPref("extensions.thumbsdown.scriptrunner.focus")
            this.util.openScriptRunnerWindow(focus);
        }
    },
    
    getScripts: function()
    {
        if (! this.scripts)
        {
            var scripts = this.sqlExec.fetchAll("getScripts");
            for each (script in scripts)
            {
                script.regexp = new RegExp("^" + script.regexp + "$");
                if (script.type == 1)
                {
                    script.isSourceThumbnail = (script.source == 0);
                    script.isSourceLink = (script.source == 1);
                    script.pattern = new RegExp("^" + script.pattern + "$");
                }
            }
            this.scripts = scripts;
        }
        return this.scripts;
    },
    
    /**
     * Retrieves Thumbs Down Context object for browser.
     *
     * @param browser 
     * browser from where the context is looked up. 
     * If null, defaults to current tab.
     * @param create if true and context doesn't exist, creates a new
     * @param persist if true, persists created context
     *
     * @return context object or null, if it doesn't exist and a new was not created
     */
    _getContext: function(browser, create, persist)
    {
        if (! browser)
            browser = gBrowser.selectedBrowser;
        var ctx = browser.thumbsdown;
        if (ctx)
            return ctx;
        if (! create)
            return null;
        ctx = new ThumbsDownContext();
        if (persist)
            browser.thumbsdown = ctx;
        return ctx;
    },
    
    _checkSettings: function()
    {
        try
        {
            var dirPath = nsPreferences.copyUnicharPref("extensions.thumbsdown.downloadDir");
            if (dirPath.length == 0)
                throw "noDir";
            var dir = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsILocalFile);
            dir.initWithPath(dirPath);
            if (! (dir.exists() && dir.isDirectory()))
                throw "dirDoesNotExist";
            return true;
        }
        catch (e if (typeof(e) == "string"))
        {
            var promptSvc = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
            var title = this.util._("error.title");
            var msg = this.util._("error." + e, [dirPath]);
            promptSvc.alert(window, title, msg);
            return false;
        }
    },
    
    _getScriptCallback: function(dir)
    {
         return function(script)
         {
             var dmSvc = Cc["@thumbsdown.mozdev.org/download-manager-service;1"].getService(Ci.nsISupports).wrappedJSObject;
             dmSvc.download([{url: script.result, referer: script.referer}], dir);
         };
    },
    
    _updateUi: function(ctx, force)
    {
        var $ = this.util.$;
        
        // is update neccessary
        var isUiActive = ($("thumbsdown_statusBarPanel").getAttribute("active") == "true");
        if (! (force || ctx.isActive || isUiActive))
            return;
        
        // status bar visibility
        var showStatusBar = nsPreferences.getBoolPref("extensions.thumbsdown.showInStatusBar");
        showStatusBar = showStatusBar || ctx.isActive;
        
        // status bar
        $("thumbsdown_statusBarPanel").hidden = (! showStatusBar);
        if (showStatusBar)
        {
            $("thumbsdown_statusBarPanel").setAttribute("active", ctx.isActive);
            $("thumbsdown_statusBarTooltip").setAttribute("active", ctx.isActive);
            if (ctx.isActive)
                $("thumbsdown_statusBarLabel").setAttribute("value", "" + ctx.thumbnails.length);
        }
        
        // commands
        $("thumbsdown_cmd_activate").setAttribute("disabled", ctx.isActive);
        $("thumbsdown_cmd_deactivate").setAttribute("disabled", (! ctx.isActive));
        $("thumbsdown_cmd_selectAll").setAttribute("disabled", (! ctx.isActive));
        $("thumbsdown_cmd_download").setAttribute("disabled", (! ctx.isActive));
    },
    
    _getBrowserForDocument: function(doc)
    {
        var browsers = gBrowser.browsers;
        for (var i = 0; i < browsers.length; i++)
            if (browsers[i].contentDocument == doc)
                return browsers[i];
        throw "browser not found"
    }
}


function ThumbsDownContext()
{
    this._isActive = false;
    this.thumbnails = [];
}

ThumbsDownContext.prototype =
{
    get isActive() { return this._isActive; },
    set isActive(value)
    {
        if (value)
            this._isActive = true;
        else
        {
            for each (image in this.thumbnails)
                this._markThumbnail(image, false);
            this._isActive = false;
            this.thumbnails = [];
        }
    },
    
    toggleThumbnail: function(image)
    {
        var index = this.thumbnails.indexOf(image);
        var selected = (index != -1);
        if (selected)
            this.deselectThumbnail(index);
        else
            selected = this.selectThumbnails([image]);
        return selected;
    },
    
    selectThumbnails: function(images)
    {
        var scripts = gThumbsDown.getScripts();
        var selected = [];
        var script = null;
        // images can be HTMLCollection (document.images) -> using normal for loop
        for (var i = 0; i < images.length; i++)
        {
            var image = images[i];
            if (this.thumbnails.indexOf(image) == -1)
            {
                script = this._isThumbnail(image, script);
                if (script)
                {
                    selected.push(image);
                    this._markThumbnail(image, true);
                    image.thumbsdown = script;
                }
            }
        }
        this.thumbnails = this.thumbnails.concat(selected);
        return (selected.length > 0);
    },
    
    deselectThumbnail: function(index)
    {
        var image = this.thumbnails.splice(index, 1)[0];
        this._markThumbnail(image, false);
    },
    
    _isThumbnail: function(image, script)
    {
        var anchor = image.parentNode;
        if (! (anchor && (anchor instanceof HTMLAnchorElement)))
            return null;
        if (script)
            if (this._matchScript(script, image, anchor))
                return script;
        var scripts = gThumbsDown.getScripts();
        for each (script in scripts)
            if (this._matchScript(script, image, anchor))
                return script;
        return null;
    },
    
    _matchScript: function(script, image, anchor)
    {
        if (! script.regexp.test(anchor.href))
            return false;
        switch (script.type)
        {
            case 0: return true;
            case 1: return script.pattern.test(script.isSourceThumbnail ? image.src : anchor.href);
            case 2: return true;
            default:
                throw new Error("invalid script type: " + script.type);
        }
    },
    
    _markThumbnail: function(image, selected)
    {
        var classNames = image.className.split(/\s+/);
        if (selected)
            classNames.push(THUMBSDOWN_CSS_SELECTED);
        else
            classNames = [x for each (x in classNames) if (x != THUMBSDOWN_CSS_SELECTED)];
        image.className = classNames.join(' ');
        if (! selected)
            image.thumbsdown = undefined; // delete is not allowed
    }
}


window.addEventListener("load", function(e) { gThumbsDown.onLoad(e); }, false);
window.addEventListener("unload", function(e) { gThumbsDown.onUnload(e); }, false);
